## Copyright © 2023,2024 Todd Coram, todd@maplefish.com, USA.

## Permission is hereby granted, free of charge, to any person obtaining
## a copy of this software and associated documentation files {the
## "Software"}, to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:

## The above copyright notice and this permission notice shall be
## included in all copies or substantial portions of the Software.

## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
## NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
## LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
## OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
## WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

## A very minimal (just enough to work for me) fastcgi protocol implementation.
##
## https://www.mit.edu/~yandros/doc/specs/fcgi-spec.html

(setq
 *FCGI_BEGIN_REQUEST       1
 *FCGI_ABORT_REQUEST       2
 *FCGI_END_REQUEST         3
 *FCGI_PARAMS              4
 *FCGI_STDIN               5
 *FCGI_STDOUT              6
 *FCGI_STDERR              7
 *FCGI_DATA                8
 *FCGI_GET_VALUES          9
 *FCGI_GET_VALUES_RESULT  10
 *FCGI_UNKNOWN_TYPE       11)

(setq
*FCGI_REQUEST_COMPLETE 0
*FCGI_CANT_MPX_CONN    1
*FCGI_OVERLOADED       2
*FCGI_UNKNOWN_ROLE     3)

(setq
 *FCGI_KEEP_CONN  1
 *FCGI_MAX_CONNS  "FCGI_MAX_CONNS"
 *FCGI_MAX_REQS   "FCGI_MAX_REQS"
 *FCGI_MPXS_CONNS "FCGI_MPXS_CONNS")

(setq
 *FCGI_RESPONDER  1
 *FCGI_AUTHORIZER 2
 *FCGI_FILTER     3
 *FCGI_RESPONDER  1
 *FCGI_AUTHORIZER 2
 *FCGI_FILTER     3)

(de _fcgi-wr2 (N)
    (wr (& 255 (>> 8 N)))
    (wr (& 255 N)))

(de fcgi-wr-record (Type ReqId D)
    (wr 1)
    (wr Type)
    (_fcgi-wr2 ReqId)
    (_fcgi-wr2 (length D))
    (wr 0)
    (wr 0)
    (mapc '((B) (if (str? B) (wr (char B)) (wr B)))  D))

(de fcgi-rd-record ()
    (let (Vers (rd 1)
	  Type (rd 1)
	  ReqId (rd 2)
	  Len (rd 2)
	  PadL (rd 1)
	  Res (rd 1)
	  Data Nil)
      (for N Len
	   (push 'Data (rd 1)))
      (for N PadL (rd 1))
      (list Type ReqId (reverse Data))))

(de fcgi-bytes-to-string (L)
    (pack (mapcar char L)))

(de fcgi-decode-nvp (Data Dpos)
    (let (Nlb (get Data Dpos)
	  Nl (if (> Nlb 127)
		 (\| (>> -24 (\& 127 Nlb)) (>> -16 (get Data (inc 'Dpos)))
		  (>> -8 (get Data (inc 'Dpos))) (get Data (inc 'Dpos)))
	       Nlb)
	  Vlb (get Data (inc 'Dpos))
	  Vl (if (> Vlb 127)
		 (\| (>> -24 (\& 127 Vlb)) (>> -16 (get Data (inc 'Dpos)))
		  (>> -8 (get Data (inc 'Dpos))) (get Data (inc 'Dpos)))
	       Vlb)
	  Nd (fcgi-bytes-to-string (head Nl (nth Data (inc 'Dpos))))
	  _ (inc 'Dpos Nl)
	  Vd  (fcgi-bytes-to-string (head Vl (nth Data Dpos))))
      (inc 'Dpos Vl)
      (list Dpos Nd Vd)))

(setq *Fcgi-params NIL)
(de fcgi-params-completed (ReqId)
    (msg "PARAMS Completed!!!"))

(de fcgi-parse-params (ReqId D)
    (let (Idx 1
	  Parms NIL)
	 (while (< Idx (length D))
	   (let (Res (fcgi-decode-nvp D Idx)
		     NewIdx (\; Res 1)
		     Name (\; Res 2)
		     Value (\; Res 3))
	     (push 'Parms (cons Name Value))
##	     (msg Name " -> " Value)
	     (setq Idx NewIdx)))
	 (if (> (length D) 0)
	     (conc (get '*Fcgi-params ReqId) Parms)
	   (fcgi-params-completed ReqId))))

(setq *Keep-open NIL)
(de fcgi-parse-begin-request (ReqId D)
    (msg "*BEGIN REQUEST")
    (put '*Fcgi-params ReqId (list (list)))
    (let (Role (\| (>> -8 (get D 1)) (get D 2))
	  Flags (get D 3))
      (setq *Keep-open (= 0 (\& Flags *FGI_KEEP_CONN)))
      (msg "  Keep-open:" (if *Keep-open "TRUE" "FALSE"))))

(de fcgi-parse-stdin (ReqId D)
    (msg "*STDIN: " (pack D)))

(de fcgi-handle-record (R)
    (let (Type (car R)
	  ReqId (cadr R)
	  D (caddr R))
      (case Type
	    (`*FCGI_BEGIN_REQUEST (fcgi-parse-begin-request ReqId D))
	    (`*FCGI_PARAMS (fcgi-parse-params ReqId D))
	    (`*FCGI_STDIN (fcgi-parse-stdin ReqId D))
	    (T (msg "Not handling " Type)))
      (and Type (list Type ReqId))))


(de fcgi-stdout (ReqId D)
    (fcgi-wr-record *FCGI_STDOUT ReqId D))

(de fcgi-end-request (ReqId Pstat)
    (default Pstat 0)
    (fcgi-wr-record *FCGI_END_REQUEST ReqId '(0 0 0 0 Pstat 0 0 0)))

(de fcgi-test ()
    (in "fastcgi-in.dat"
	(loop
	 (NIL (fcgi-handle-record (fcgi-rd-record))))))


