## Copyright © 2023,2024 Todd Coram, todd@maplefish.com, USA.

## Permission is hereby granted, free of charge, to any person obtaining
## a copy of this software and associated documentation files {the
## "Software"}, to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:

## The above copyright notice and this permission notice shall be
## included in all copies or substantial portions of the Software.

## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
## NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
## LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
## OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
## WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(load "geo-strips.l")
(load "fastcgi.l")

(de nearby-as-json (My-Lat My-Lon Km Cats)
    (let (Res (bench (geo-nearby My-Lat My-Lon Km Cats))
	  JList Nil)
      (push 'JList (pack "{ \"count\" : " (length Res)))
      (push 'JList (pack ", \"items\" : \["))
      (for H Res
	   (let (Dkm (\; H 1)
		 Name (\; H 2)
		 House (\; H 3)
		 Street (\; H  4)
		 City (\; H 5)
		 Lat (\; H 7)
		 Lon (\; H 8)
		 Miles (geo-km-to-mile Dkm))
	     (push 'JList (text "{ \"name\" : \"@1\", \"house\" : \"@2\",\
				\"street\" : \"@3\", \"city\" : \"@4\", \
				\"miles\" : @5, \"lat\" : @6, \"lon\" : @7 }"
				Name House Street City (format Miles *Scl)
				(format Lat *Scl) (format Lon *Scl)))))
      JList))

(de nearby-as-html (My-Lat My-Lon Km Cats)
    (let (Res (bench (geo-nearby My-Lat My-Lon Km Cats))
	  HList "")
      (push 'HList (pack "<p>Found " (length Res) " hits...</p>"))
      (for H Res
	   (let (Dkm (\; H 1)
		 Name (\; H 2)
		 House (\; H 3)
		 Street (\; H  4)
		 City (\; H 5)
		 Lat (\; H 7)
		 Lon (\; H 8)
		 Href (text "<a href=\"https://maps.google.com/?z=12&q=@1,@2\">@3</a>"
			    (format Lat *Scl) (format Lon *Scl)  Name)
		 Miles (geo-km-to-mile Dkm)
		 Dist (if (< Miles 0.25)
			  (pack (round (* Miles 5280) 0) " feet away")
			(pack (round Miles 2) " miles away")))
	     (push 'HList (text "<p>@1 @2 @3 @4 (@5)</p>"
			       Href House Street City Dist)))) HList))

(de fcgi-params-completed (R)
    (let Qs (cadr (assoc "QUERY_STRING" (get '*Fcgi-params R)))
	 (msg "Working on Query -> " Qs)
	 (for Pair (split (chop Qs) "&")
	      (let P (split (chop Pair) "=")
		   (case (pack (car P))
			 ("lat" (msg "Lat=" (setq Lat (format (cadr P) *Scl))))
			 ("lon" (msg "Lon=" (setq Lon (format (cadr P) *Scl))))
			 ("miles" (msg "Radius=" (setq Miles (format (cadr P) *Scl))))
			 ("cats" (msg "Cats=" (setq Cats (split (cadr P) ":"))))))))
    (setq Cats (delete NIL Cats))
    (out *Socket
	 (let (Res
	       (bench (nearby-as-html Lat Lon
				      (geo-mile-to-km (* 1.0 Miles)) (mapcar pack Cats))))
	   (msg "Hits: " (length Res))

	   (fcgi-stdout R (chop "Content-type: text/html\r\n"))
	   (fcgi-stdout R (chop "Access-Control-Allow-Origin: *\r\n\r\n<html>"))
	   (fcgi-stdout R (chop (pack (reverse Res))))
	   (fcgi-stdout R (chop "</html>\r\n"))
	   (fcgi-stdout R (chop "")))
	 (fcgi-end-request R)))

(de loca-server (P)
    (let Port (port P)
	 (loop
	  (gc)
	  (msg "Waiting for a connection on " P)
	  (setq *Socket (listen Port))
	  (msg "Got a connection!")
	  (in *Socket
	      (loop
	       (NIL (fcgi-handle-record (fcgi-rd-record)))))
	  (close *Socket))))


(de run-loco ()
    (geo-init)
    (msg "Loading database...")
    (geo-load-native "conus.dat")
    (loop
     (catch T (finally (close *Socket)
		       (loca-server 6666)))))

(run-loco)


